import numpy as np
from sklearn.utils.extmath import safe_sparse_dot

def inner_product(X, Y):
    """
    Arguments
    --------
    X, Y: numpy.ndarray or scipy.sparse.csr_matrix
        One of both must be sparse matrix
        shape of X = (n,p)
        shape of Y = (p,m)

    Returns
    -------
    Z : scipy.sparse.csr_matrix
        shape of Z = (n,m)
    """

    return safe_sparse_dot(X, Y, dense_output=False)

def check_sparsity(x):
    """
    It calculates sparsity of centroid vectors

    Argument
    --------
    x : numpy.ndarray

    Returns
    -------
    sparsity : float
        1 - proportion of nonzero elements
    """
    n,m = x.shape
    return sum(len(np.where(x[c] != 0)[0]) for c in range(n)) / (n*m)




import numpy as np
from sklearn.metrics import pairwise_distances

def merge_close_clusters(centers, labels, max_dist=0.7):
    """
    Arguments
    ---------
    centers : numpy.ndarray
        Shape = (k, p)
        k : num of clusters
        p : num of features
    labels : numpy.ndarray or None
        Shape = (n,)
        n : num of documents in the dataset
        It is used as cluster size weight. If the value is None,
        it assumes that all size of clusters are equal.
    max_dist : float
        Maximum Cosine distance between base cluster and other clusters
        The smaller the value, it groups closer clusters to a group.

    Returns
    -------
    group_centers : numpy.ndarray
        Merged cluster centroid vectors
        Shape = (l, p)
        l : num of merged clusters. l <= k
        p : num of features
    groups : list of list of int
        The index of first index corresponds the index of merged clusters.
        And the value in the inner lists corresponds the index of original clusters.
    """

    n_clusters, n_terms = centers.shape
    cluster_size = np.bincount(labels, minlength=n_clusters)
    sorted_indices, _ = zip(*sorted(enumerate(cluster_size), key=lambda x:-x[1]))

    groups = _grouping_with_centers(centers, max_dist, sorted_indices)
    centers_ = np.dot(np.diag(cluster_size), centers)

    n_groups = len(groups)
    group_centers = np.zeros((n_groups, n_terms))
    for g, idxs in enumerate(groups):
        sum_ = centers_[idxs].sum(axis=0)
        mean = sum_ / cluster_size[idxs].sum()
        group_centers[g] = mean
    return group_centers, groups

def _closest_group(groups, c, pdist, max_dist):
    """
    It finds the most closest merged clusters with the given cluster c

    Arguments
    ---------
    groups : list of list of int
        The index of first index corresponds the index of merged clusters.
        And the value in the inner lists corresponds the index of original clusters.
    c : int
        Index of the cluster which this function finds the most closest merged group
    pdist : numpy.ndarray
        Pairwise distance matrix.
        Shape is (k, k) where k is num of clusters.
    max_dist : float
        Maximum Cosine distance between base cluster and other clusters.
        The smaller the value, it groups closer clusters to a group.

    Returns
    -------
    closest : int or None
        The index of most closest merged group.
        If the closest one and target cluster c is distants than max_dist,
        it returns None
    """

    dist_ = 1
    closest = None
    for g, idxs in enumerate(groups):
        dist = pdist[idxs, c].mean()
        if dist > max_dist:
            continue
        if dist_ > dist:
            dist_ = dist
            closest = g
    return closest

def _grouping_with_centers(centers, max_dist, sorted_indices):
    """
    Arguments
    ---------
    centers : numpy.ndarray
        Shape = (k, p)
        k : num of clusters
        p : num of features
    max_dist : float
        Maximum Cosine distance between base cluster and other clusters.
        The smaller the value, it groups closer clusters to a group.
    sorted_index : list of int
        Cluster index sorted by the size of each cluster.

    Returns
    -------
    groups : list of list of int
        The index of first index corresponds the index of merged clusters.
        And the value in the inner lists corresponds the index of original clusters.
    """

    pdist = pairwise_distances(centers, metric='cosine')
    return _grouping_with_pdist(pdist, max_dist, sorted_indices)

def _grouping_with_pdist(pdist, max_dist, sorted_indices):
    """
    Arguments
    ---------
    pdist : numpy.ndarray
        Pairwise distance matrix.
        Shape is (k, k) where k is num of clusters.
    max_dist : float
        Maximum Cosine distance between base cluster and other clusters.
        The smaller the value, it groups closer clusters to a group.
    sorted_index : list of int
        Cluster index sorted by the size of each cluster.

    Returns
    -------
    groups : list of list of int
        The index of first index corresponds the index of merged clusters.
        And the value in the inner lists corresponds the index of original clusters.
    """
    groups = [[sorted_indices[0]]]
    for c in sorted_indices[1:]:
        g = _closest_group(groups, c, pdist, max_dist)
        # create new group
        if g is None:
            groups.append([c])
        # assign c to g
        else:
            groups[g].append(c)
    return groups



import numpy as np


def proportion_keywords(centers, labels=None, min_score=0.5, topk=200,
                        candidates_topk=300, index2word=None, stopwords=None):
    """
    Arguments
    ---------
    centers : numpy.ndarray
        Shape = (k, p)
        k : num of clusters
        p : num of features
    labels : numpy.ndarray or None
        Shape = (n,)
        n : num of documents in the dataset
        It is used as cluster size weight. If the value is None,
        it assumes that all size of clusters are equal.
    min_score : float
        0 < min_score < 1
        Minimum keyword score. The words corresponding less than the score will be
        removed from set of keywords.
    topk : int
        The number of keywords
    candidates_topk : int
        The number of keyword candidates
    index2word : list of str or None
        If the value is not None, the length must be p (num of features)
        If the value is None, it returns keywords as integer index
    stopwords : set of str or None
        The set of words that will be removed from keyword set by force

    Returns
    -------
    keywords_in_clusters : nested list
        The length of list equals with num of clusters
        [[(keyword, score), (keyword, score), ... ],
         [(keyword, score), (keyword, score), ... ]
         ...
        ]
    """

    def l1_normalize(x): return x/x.sum()

    password_as_idx = stopwords and isinstance(list(stopwords)[0], int)

    n_clusters, n_features = centers.shape
    total_frequency = np.zeros(n_features)

    if labels is not None:
        n_samples_in_cluster = np.bincount(labels, minlength=n_clusters)
    else:
        n_samples_in_cluster = np.asarray([1] * n_clusters)

    for c, n_docs in enumerate(n_samples_in_cluster):
        total_frequency += (centers[c] * n_docs)
    total_sum = total_frequency.sum()

    keywords = []
    for c, n_docs in enumerate(n_samples_in_cluster):
        if n_docs == 0:
            keywords.append([])
            continue

        n_prop = l1_normalize(total_frequency - (centers[c] * n_docs))
        p_prop = l1_normalize(centers[c])

        indices = np.where(p_prop > 0)[0]
        indices = sorted(indices, key=lambda idx: -
                         p_prop[idx])[:candidates_topk]
        keywords_c = [(idx, p_prop[idx] / (p_prop[idx] + n_prop[idx]))
                  for idx in indices]
        keywords_c= [t for t in keywords_c if t[1] >= min_score]
        keywords_c = sorted(keywords_c, key=lambda x: -x[1])

        if stopwords and password_as_idx:
            keywords_c = [t for t in keywords_c if t[0] not in stopwords]

        if index2word is not None:
            keywords_c = [(index2word[idx], score) for idx, score in keywords_c]
            if stopwords and not password_as_idx:
                keywords_c = [t for t in keywords_c if t[0] not in stopwords]

        keywords.append(keywords_c)


    if topk > 0:
        keywords = [keyword[:topk] for keyword in keywords]

    return keywords



from sklearn.metrics import pairwise_distances
import matplotlib.pyplot as plt
import numpy as np

from ._postprocess import _grouping_with_pdist

def visualize_pairwise_distance(centers, labels=None, max_dist=0.7, 
    sort=False, show=True, title=None, figsize=(15,15), cmap='gray',
    clim=None, dpi=50, facecolor=None, edgecolor=None, frameon=True):
    """
    Arguments
    ---------
    centers : numpy.ndarray
        Shape = (k, p)
        k : num of clusters
        p : num of features
    labels : numpy.ndarray or None
        Shape = (n,)
        n : num of documents in the dataset
        It is used as cluster size weight. If the value is None,
        it assumes that all size of clusters are equal.
    max_dist : float
        Maximum Cosine distance between base cluster and other clusters.
        The smaller the value, it groups closer clusters to a group.
    sort : Boolean
        It True, it first groups nearby clusters into one group
        then draw pairwise distance matrix
    title : str or None
        The title of the figure
    figsize : tuple of int
        Size of the figure. Default is (15,15)
    cmap : str
        Color map in matplotlib. Default is gray
    clim : tuple of int or None
        Color limit in matplotlib. 
    dpi : int
        DPI of the figure
    facecolor : tuple of float or None
        RBG color of background
        For example (0.15, 0.5, 0.5)
    edgecolor : tuple of float or None
        RBG color of edge for the figure
        For example (0.15, 0.5, 0.5)
    frameon : Boolean
        Ignored

    Returns
    -------
    figure : matplotlib.pyplot.figure
        Figure of pairwise distance
    """

    n_clusters = centers.shape[0]

    if labels is None:
        labels = [i for i in range(n_clusters)]

    pdist = pairwise_distances(centers, metric='cosine')

    if sort:
        # sort clusters by cluster size
        cluster_size = np.bincount(labels, minlength=n_clusters)
        sorted_indices, _ = zip(*sorted(enumerate(cluster_size), key=lambda x:-x[1]))

        # grouping clusters
        groups = _grouping_with_pdist(pdist, max_dist, sorted_indices)

        # sort groups by group size
        groups = sorted(groups, key=lambda x:-len(x))

        # create revised index
        sorted_indices = [idx for group in groups for idx in group]
        indices_revised = np.ix_(sorted_indices, sorted_indices)

        # create origin index
        indices_orig = list(range(n_clusters))

        # revise pairwise distance matrix
        pdist_revised = np.empty_like(pdist)
        pdist_revised[np.ix_(indices_orig,indices_orig)] = pdist[indices_revised]

    else:
        pdist_revised = pdist

    figure = pairwise_distance_to_matplotlib_figure(pdist_revised, title,
        figsize, cmap, clim, dpi, facecolor, edgecolor, frameon)
    return figure

def pairwise_distance_to_matplotlib_figure(pdist, title=None,
    figsize=(15,15), cmap='gray', clim=None, dpi=50,
    facecolor=None, edgecolor=None, frameon=True):
    """
    pdist : numpy.ndarray
        Pairwise distance matrix. Shape is (k, k) when k is num of clusters
    title : str or None
        The title of the figure
    figsize : tuple of int
        Size of the figure. Default is (15,15)
    cmap : str
        Color map in matplotlib. Default is gray
    clim : tuple of int or None
        Color limit in matplotlib. 
    dpi : int
        DPI of the figure
    facecolor : tuple of float or None
        RBG color of background
        For example (0.15, 0.5, 0.5)
    edgecolor : tuple of float or None
        RBG color of edge for the figure
        For example (0.15, 0.5, 0.5)
    frameon : Boolean
        Ignored

    Returns
    -------
    figure : matplotlib.pyplot.figure
        The figure of pairwise distance matrix
    """

    n_clusters = pdist.shape[0]

    figure = plt.figure(
        figsize = figsize,
        dpi = dpi,
        facecolor = facecolor,
        edgecolor = edgecolor,
        frameon = frameon
    )

    if title:
        plt.title(title)

    plt.xlim((0, n_clusters))
    plt.ylim((0, n_clusters))

    if clim:
        plt.imshow(pdist, cmap=cmap, clim=clim)
    else:
        plt.imshow(pdist, cmap=cmap)

    return figure



import soyclustering
from setuptools import setup, find_packages

with open('README.md', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name="soyclustering",
    version=soyclustering.__version__,
    author=soyclustering.__author__,
    author_email='soy.lovit@gmail.com',
    url='https://github.com/lovit/clustering4docs',
    description="Python library for document clustering",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=["numpy>=1.1"],
    keywords = ['document clustering', 'clustering labeling'],
    packages=find_packages()
)



__title__ = 'soyclustering'
__version__ = '0.2.0'
__author__ = 'Lovit'
__license__ = 'GPL v3'
__copyright__ = 'Copyright 2017 Lovit'

from ._kmeans import SphericalKMeans
from ._keyword import proportion_keywords
from ._postprocess import merge_close_clusters
from ._visualize import visualize_pairwise_distance
from ._visualize import pairwise_distance_to_matplotlib_figure
